// -*- C++ -*-
// $RCSfile: equation.C,v $
// $Revision: 1.77.8.1 $
// $Author: langer $
// $Date: 2010-05-13 20:57:16 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/switchboard.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "engine/csubproblem.h"
#include "engine/edge.h"
#include "engine/element.h"
#include "engine/equation.h"
#include "engine/field.h"
#include "engine/fieldeqnlist.h"
#include "engine/flux.h"
#include "engine/fluxnormal.h"
#include "engine/gausspoint.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/nodalequation.h"
#include "engine/node.h"
#include "engine/ooferror.h"
#include "engine/property.h"
#include "timederivative.h"

std::vector<Equation*> &Equation::all() {
  static std::vector<Equation*> all_eqns;
  return all_eqns;
}

Equation *getEquationByIndex(int index) {
  Equation *eqn = Equation::all()[index];
  return eqn;
}

Equation *Equation::getEquation(const std::string &nm) {
  const std::vector<Equation*> &list = all();
  for(std::vector<Equation*>::size_type i=0; i<list.size(); i++) {
    if(list[i]->name() == nm) {
      return list[i];
    }
  }
  throw ErrProgrammingError("Unknown Equation \"" + nm + "\"",
			    __FILE__, __LINE__);
}

FEWrapper<Equation>::AllWrappers
&Equation::FindAllEquationWrappers::operator()()
{
    return mesh->equationwrappers;
}

int countEquations() {
  return Equation::all().size();
}

const std::string Equation::modulename_("ooflib.SWIG.engine.equation");

Equation::Equation(const std::string &nm, Flux &flx, int d)
  : kinetic(0),
    dynamic(0),
    kinetic_activity(false),
    dynamic_activity(false),
    timeDerivativeComponents(0),
    dim_(d),
    name_(nm),
    fflux(&flx),
    index_(all().size())
{
  all().push_back(this);
  fflux->addEquation(this); 
}

// position of a given component in the eqn lists in the nodes

int Equation::localindex(const FuncNode &node, const FieldIndex &component)
  const
{
  return node.equationset.offset(this) + component.integer();
}

int Equation::localindex(const FuncNode &node, int component) const {
  return node.equationset.offset(this) + component;
}

NodalEquation *Equation::nodaleqn(const FuncNode &node, int component) const {
  return node.eqnlist[localindex(node, component)];
}

// activate_fluxes() and deactivate_fluxes() are called by
// CSubProblem::activate_equation() and // CSubProblem::deactivate_equation().

void Equation::activate_fluxes(CSubProblem *subpr) {
//  Trace("Equation::activate " + name_);
  subpr->activate_flux(*fflux);
}

void Equation::deactivate_fluxes(CSubProblem *subpr) {
  subpr->deactivate_flux(*fflux);
}

bool Equation::is_active(const CSubProblem *subpr) const {
  return subpr->is_active_equation(*this);
}

const std::string &Equation::fluxname() const {
  return fflux->name();
}

const Flux *Equation::flux() const {
  return fflux;
}


void Equation::addTimeDerivative(TimeDerivative *td)
{
  timeDerivativeComponents.push_back(td);
}

const std::vector<TimeDerivative*> &Equation::allTimeDerivatives()
{
  return timeDerivativeComponents;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

DivergenceEquation::DivergenceEquation(const std::string &name, Flux &flx,
				       int d)
  : Equation(name, flx, d)
{}

const std::string &DivergenceEquation::classname() const {
  static std::string name_("DivergenceEquation");
  return name_;
}

// Equation::make_stiffness is called by Flux::make_stiffness after
// the Properties have computed the flux matrix for the Flux at the
// current GaussPoint.  Equation::make_stiffness has to use the flux
// matrix to compute the GaussPoints contribution to the global
// stiffness matrix.

void DivergenceEquation::make_stiffness(CSubProblem *subp, const Element *el,
					const GaussPoint &gpt,
					const Flux &flux,
					const FluxData *fluxdata,
					LinearizedSystem &system)
  const
{
  SparseLinkMatrix<double> &mtx = system.get_stiffness();

//  Trace("DivergenceEquation::make_stiffness " + name());
  if(flux.divergence_dim() != dim()) {
    throw ErrProgrammingError("Divergence has wrong dimension",
			      __FILE__, __LINE__);
  }
  double wgt = gpt.weight();
  MATRIX_D localK(dim(), el->ndof()); // local stiffness
  
  // loop over nodes.  These are the test functions.
  for(ElementFuncNodeIterator node=el->funcnode_iterator();!node.end();++node) {
    localK.set(0.0);

    flux.local_stiffness(el, node, gpt, fluxdata, localK); // compute localK

    // loop over components of equation (e.g., components of flux divergence)
    for(int eqcomp=0; eqcomp<dim(); eqcomp++) {
      // eqcomp corresponds to an index that runs from 0 to the number
      // of equation components in the Equation object, and NOT to the
      // NodalEquation component in the node.

      // Row in global matrix, symmetrized if possible.
      int rowno = subp->map_forward_index(*this, *node.funcnode(), eqcomp);
      
      // This row is an in-plane equation.
      system.rowgroup[rowno] = -1;

      // Loop over degrees of freedom of the element that couple to
      // this flux, by looping over all components of the relevant
      // *defined* fields at all the nodes.  The out-of-plane fields
      // are included in the field list.
      const std::vector<const Field*> &fieldlist =
	el->material()->fieldlist(flux);
      for(std::vector<const Field*>::size_type f=0; f<fieldlist.size(); f++) {
	const Field *field = fieldlist[f];
	if(subp->is_defined_field(*field)) {
 	  // Trace("DivergenceEquation::make_stiffness: field " + field->name()
	  //	+ " is defined");
	  IteratorP fi = field->iterator(ALL_INDICES); // reuse this!
	  for(ElementFuncNodeIterator nj=el->funcnode_iterator();!nj.end();++nj)
	    {
	      // loop over components of field
	      const FuncNode *nodej = nj.funcnode();
	      for(; !fi.end(); ++fi) {
		int localcolno = nj.localindex(*field, fi);
      		DegreeOfFreedom *dof = (*field)(nodej, fi.integer());
		// The integrals over the element area are accumulated
		// directly in the stiffness matrix.  The gauss point weight
		// is finally included here.
		double raw = localK(eqcomp, localcolno);
		double res = raw * wgt;

//	 	Trace("DivergenceEquation adding " + to_string(res) + 
// 		      " to stiffness matrix (" + to_string(rowno) + ", " + 
// 		      to_string(dof->dofindex()) + "),   Integrand is " + 
// 		      to_string(raw) + ", weight is " + to_string(wgt));

		mtx(rowno, dof->dofindex()) += res;

		// Mark this column as either in-plane or part of a group.
		system.colgroup[dof->dofindex()] = 
		  (field->in_plane_part() ? -1 : nj.node()->index() );

	      }	// end loop over field components
	      fi.reset();
	    } // end loop over nodes
	} // end if field is defined
      } // end loop over fields
    } // end loop over equation components
  } // end loop over test functions
}

int DivergenceEquation::integration_order(const Element *el) const {
  return el->dshapefun_degree();
}

// At this point, we have some force values for the RHS in the flux's
// fluxrhs vector.  Need to distribute them.  Called by
// Flux::make_rhs, which is called by Material::make_linear_system
// while integrating over gauss points within an element.
void DivergenceEquation::make_rhs(CSubProblem *subp, const Element *element, 
				  const GaussPoint &gpt, const Flux &flux,
				  const FluxData *fluxdata, 
				  LinearizedSystem  &system)
  const {

  VECTOR_D &rhs = system.get_body_rhs();

  VECTOR_D rhs_from_offset(dim());

  // For each component of the equation, for each node in the element,
  // contribute to the appropriate nodal equation.
  for(ElementFuncNodeIterator node=element->funcnode_iterator();
      !node.end();++node) {

    // Call back to the flux, in case it has contributions from 
    // the flux offset vector.
    flux.make_offset(element, node, gpt, fluxdata, rhs_from_offset);
    double sfvalue = node.shapefunction(gpt);
    for(int eqcomp=0; eqcomp<dim(); eqcomp++) {
      // Put the RHS contribution into the subp's rhs object here.
      // The argument of rhs corresponds to the row number 
      // of the master stiffness matrix
      int nrow = subp->map_forward_index(*this, 
					 *node.funcnode(), 
					 eqcomp); // symmetrized row
      rhs[nrow] +=
	- (fluxdata->rhs_element(eqcomp)*sfvalue*gpt.weight() // body forces
	   + rhs_from_offset[eqcomp]*gpt.weight() ); // fixed "fluxes"
    }
  }
}

// Boundary integral for a divergence equation.  Calls back to the
// calling flux to fill the rhs structure, then distributes the
// results, multiplied by the gauss weight, to the nodal equation
// value fields.  
void DivergenceEquation::boundary_integral(const CSubProblem *subp,
					   const Flux *flux,
					   const BoundaryEdge *ed, 
					   const EdgeGaussPoint& egp, 
					   const FluxNormal *flxnormal) 
  const {
  
  VECTOR_D &brhs(subp->boundary_rhs());
  std::vector<double> J_dot_normal(dim(),0.0);
  for(EdgeNodeIterator edi = ed->node_iterator(); !edi.end(); ++edi) {
    // Call the flux to populate it.
    flux->local_boundary(ed, edi, egp, flxnormal, J_dot_normal);
    // And distribute the values to the boundary rhs vector.
    for(std::vector<double>::size_type i=0; i<J_dot_normal.size(); i++) {
      int rownum = subp->map_forward_index(*this, *edi.funcnode(), i);
      brhs[rownum] += egp.weight()*J_dot_normal[i];
    }
  }
}

IteratorP DivergenceEquation::iterator() const {
  return fflux->divergence_iterator();
}

IndexP DivergenceEquation::componenttype() const {
  return fflux->divergence_componenttype();
}

IndexP DivergenceEquation::getIndex(const std::string &str) const {
  return fflux->divergence_getIndex(str);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

PlaneFluxEquation::PlaneFluxEquation(const std::string &name, Flux &flx, int d)
  : Equation(name, flx, d)
{}

const std::string &PlaneFluxEquation::classname() const {
  static const std::string name_("PlaneFluxEquation");
  return name_;
}

void PlaneFluxEquation::boundary_integral(const CSubProblem *m,
					  const Flux *flux,
					  const BoundaryEdge *ed,
					  const EdgeGaussPoint& egp,
					  const FluxNormal *flxnormal) 
  const {
  
  return; // stub
}

void PlaneFluxEquation::make_stiffness(CSubProblem *subp, const Element *el,
				       const GaussPoint &gpt,
				       const Flux &flux,
				       const FluxData *fluxdata,
				       LinearizedSystem &system)
  const
{
  double wgt = gpt.weight();

  SparseLinkMatrix<double> &mtx = system.get_stiffness();
 
  // loop over nodes
  for(ElementFuncNodeIterator node=el->funcnode_iterator(); !node.end(); ++node)
    {
      double shapefun = node.shapefunction(gpt);
      // loop over components of equation (ie,  of out-of-plane flux)	
      int eqcomp = 0;
      for(IteratorP fluxiter=flux.iterator(OUT_OF_PLANE);
	  eqcomp<dim() && !fluxiter.end(); ++eqcomp, ++fluxiter) {
	const FieldIndex &fluxindex = fluxiter;
	
	int rowno = subp->map_forward_index(*this, 
					    *node.funcnode(), 
					    eqcomp); //symmetric row component
	// This row is an out-of-plane equation.
	system.rowgroup[rowno] = node.node()->index();

	// loop over degrees of freedom of the element that couple to
	// this flux, by looping over all components of the relevant
	// *defined* fields at all the nodes.  The out-of-plane fields
	// are included in the field list.
	const std::vector<const Field*> &fieldlist = 
	  el->material()->fieldlist(flux);
	for(std::vector<const Field*>::size_type f=0; f<fieldlist.size(); f++) {
	  const Field *field = fieldlist[f];
	  if(subp->is_defined_field(*field)) {
	    // loop over components of active Field at each Node
	    IteratorP fi = field->iterator(ALL_INDICES); // reuse this!
	    for(ElementFuncNodeIterator nj=el->funcnode_iterator();
		!nj.end(); ++nj)
	      {
		const FuncNode *nodej = nj.funcnode();
		for(; !fi.end(); ++fi) {
		  const FieldIndex &fidx = fi; //convert IteratorP to FieldIndex
		  DegreeOfFreedom *dof = (*field)(nodej, fidx.integer());
		  
		  double raw = fluxdata->matrix_element(fluxindex, field,
							fidx, nj) * shapefun;
		  double res = raw * wgt;

// 		  Trace("PlaneFluxEquation adding " + to_string(res) + 
// 			" to stiffness matrix (" + to_string(rowno) + ", " + 
// 			to_string(dof->dofindex()) + "),   Integrand is " + 
// 			to_string(raw) + ", weight is " + to_string(wgt));

		  mtx(rowno, dof->dofindex()) += res;

		  // Mark this column as either in-plane or part of a group.
		  system.colgroup[dof->dofindex()] = 
		    (field->in_plane_part() ? -1 : nj.node()->index() );

		} // end loop over field components
		fi.reset();
	      } // end loop over nodes
	  }
	}
      }
    }
}

int PlaneFluxEquation::integration_order(const Element *el) const {
  return el->shapefun_degree();
}

void PlaneFluxEquation::make_rhs(CSubProblem *subp, const Element *element,
				 const GaussPoint &gpt, const Flux &flux,
				 const FluxData *fluxdata, 
				 LinearizedSystem &system)
  const {
  
  VECTOR_D &rhs = system.get_body_rhs();

  // For each component, for each node, contribute to the appropriate
  // nodal equation.  Our contribution is proportional to the
  // appropriate value of the flux offset directly.
  for(ElementFuncNodeIterator node=element->funcnode_iterator();
      !node.end();++node) {
    double sfvalue = node.shapefunction(gpt);
    
    // For each out-of-plane component of the flux, make a
    // contribution to the corresponding component of this equation --
    // the contribution comes from the offset of the flux to which we
    // refer, and it's the direct offset, not the rhs, because we
    // don't have a divergence.
    int eqcomp = 0;
    for(IteratorP fluxiter=flux.iterator(OUT_OF_PLANE);
	eqcomp<dim() && !fluxiter.end(); ++eqcomp, ++fluxiter) {
      
      int nrow = subp->map_forward_index(*this,
					 *node.funcnode(),
					 eqcomp); // Row, symmetrized.
      rhs[nrow] -= 
	fluxdata->offset_element(fluxiter)*sfvalue*gpt.weight();
    }
  }
  
}


IteratorP PlaneFluxEquation::iterator() const {
  return fflux->out_of_plane_iterator();
}

IndexP PlaneFluxEquation::componenttype() const {
  return fflux->componenttype();
}

IndexP PlaneFluxEquation::getIndex(const std::string &str) const {
  return fflux->getIndex(str);
}

//////////////
 
std::ostream &operator<<(std::ostream &os, const Equation &eqn) {
  os << "Equation(" << eqn.name() << ")";
  return os;
}


