// -*- C++ -*-
// $RCSfile: material.C,v $
// $Revision: 1.130.6.2 $
// $Author: reida $
// $Date: 2008-12-15 19:52:58 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#ifdef DEBUG
#include <assert.h>
#endif
#include "common/activearea.h"
#include "common/cmicrostructure.h"
#include "common/coord.h"
#include "common/pixelattribute.h"
#include "common/IO/stringimage.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "common/printvec.h"
#include "engine/csubproblem.h"
#include "engine/elementnodeiterator.h"
#include "engine/ooferror.h"
#include "engine/property/color/color.h"
#include "engine/IO/propertyoutput.h"
#include "engine/element.h"
#include "engine/flux.h"
#include "engine/mastercoord.h"
#include "engine/material.h"
#include "engine/equation.h"
#include "engine/timederivative.h"
#include <iostream>
#include <map>

//Interface branch
Material::Material(const std::string &nm,
		   const std::string &materialtype)
  : name_(nm),
    consistency_state(0),
    type_(materialtype),
    fluxprop(Flux::allfluxes().size()),
    fluxfield(Flux::allfluxes().size()),
    outputprop(nPropertyOutputRegistrations())
{}

Material::~Material() {}

bool Material::self_consistent() const {
  return consistency_state.size()==0;
}

const std::vector<std::string> *Material::consistency_messages() const {
  return &consistency_state; 
}

const TimeStamp &Material::getTimeStamp() const {
  return timestamp;
}

std::ostream &operator<<(std::ostream &os, const Material &mat) {
  os << "Material('" << mat.name();
  for(std::vector<Property*>::size_type i=0; i<mat.property.size(); i++)
    os << ", " << mat.property[i];
  os << "')";
  return os;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

class MaterialPropertyRegistration {
private:
  const std::string tag;
  Property *property;
  MaterialPropertyRegistration(Property *p, const std::string &str)
    : tag(str), property(p)
  {}
  friend class MaterialPropertyRegistry;
};

// Register a property under the name "tag"
void MaterialPropertyRegistry::registr(Property *prop, const std::string &tag) {
  if(fetch(tag)) {		// Can't have two Properties with the same name
    throw ErrRedundantProperty(tag);
  }
  reg.push_back(new MaterialPropertyRegistration(prop, tag));
}

// Find the property registered with the name "tag"
Property *MaterialPropertyRegistry::fetch(const std::string &tag) const {
  for(std::vector<MaterialPropertyRegistration*>::size_type i=0; i<reg.size();
      i++)
    {
      if(reg[i]->tag == tag) {
	return reg[i]->property;
      }
    }
  return 0;
}

MaterialPropertyRegistry::~MaterialPropertyRegistry() {
  clear();
}

void MaterialPropertyRegistry::clear() {
  for(std::vector<MaterialPropertyRegistration*>::size_type i=0; i<reg.size();
      i++)
    delete reg[i];
  reg.clear();
}

#ifdef DEBUG
void MaterialPropertyRegistry::dump(std::ostream &os) const {
  for(std::vector<MaterialPropertyRegistration*>::size_type i=0; i<reg.size();
      i++)
    {
      os << reg[i]->property->name()
	 << " (" << reg[i]->property->classname() << ") "
	 << reg[i]->tag << std::endl;
    }
}
#endif // DEBUG

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//


Property *Material::getProperty(int i) const { 
  return property[i];
}


void Material::add1Property(Property *prop) {
  property.push_back(prop);
  ++timestamp;
}

void Material::remove1Property(Property *prop) {
  for(std::vector<Property*>::iterator p=property.begin(); p<property.end();
      ++p)
    {
      if(*p == prop) {
	property.erase(p);
	++timestamp;
	return;
      }
    }
  throw ErrNoSuchProperty(name(), prop->name());
}

void Material::registerPropertyType(Property *p, const std::string &nm) {
  // Called by Property::bookkeeping().
  registry.registr(p, nm);
}

Property *Material::fetchProperty(const std::string &nm) const {
#ifdef DEBUG
  // registry.dump(std::cerr);
#endif
  Property *prop = registry.fetch(nm);
  if(!prop) {
    throw ErrNoSuchProperty(name(), nm);
  }
  return prop;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Call cross_reference() whenever properties have been added to or
// removed from the material, so the "consistency_state" flag/object is 
// always correct.

void Material::cross_reference() {
  clear_xref();
  consistency_state.clear();  
  

  // ask properties to tell us what kind they are, what fields they use, etc.
  for(std::vector<Property*>::size_type i=0; i<property.size(); i++) {
    try {
      // Property::bookkeeping calls:
      //    Material::registerPropertyType (fills in MaterialPropertyRegistry)
      //    Field::registerProperty (which calls Property::require_field)
      //    Material::registerFlux (fills in FluxPropLists)
      property[i]->bookkeeping(this);
    }
    catch(ErrRedundantProperty &exc) {
      // The redundant-property exception is thrown from within the
      // MaterialPropertyRegistry class, when props are registering
      // themselves.  It doesn't know the material -- we could pass it
      // through, but it's easy to just add on here.
      consistency_state.push_back("Property '" + property[i]->name() + \
	"' is a redundant property of type '" + exc.get_tag() + "'.");
    }
  }

  // ask each property to use the registry to find other Properties.
  for(std::vector<Property*>::size_type i=0; i<property.size(); i++) {
    try {
      property[i]->cross_reference(this);
    }
    catch(ErrNoSuchProperty &exc) {
      consistency_state.push_back(exc.message());
    }
  }

  // Use FluxPropLists to find which Fields contribute to which Fluxes.
  make_fluxfield_table();
}


void Material::clear_fluxproplist() {
  for(std::vector<FluxPropList>::size_type i=0; i< fluxprop.size(); i++) 
    fluxprop[i].clear();
}

void Material::clear_xref() {
  clear_fluxproplist();
  clear_fluxfield_table();
  clear_outputprop();
  registry.clear();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Material::registerFlux(Property *prop, const Flux *flux) {
  // This property contributes to this flux.  Called by
  // Property::bookkeeping().
//   Trace("Material::registerFlux " + prop->name() + " " + flux->name())
  fluxprop[flux->index()].push_back(prop);
}

// Routine to query whether or not this material has properties which
// contribute to the indicated flux -- this is true if the
// corresponding fluxprop entry has nonzero length.
bool Material::contributes_to_flux(const Flux *flux) const {
  if (fluxprop[flux->index()].size() != 0)
    return true;
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Material::registerOutput(Property *prop, const std::string &outputname) {
  // The given Property, which is present in this material,
  // contributes to the given PropertyOutput.
  PropertyOutputRegistration *reg = getPropertyOutputReg(outputname);
  if(!reg)
    throw ErrProgrammingError("PropertyOutputRegistration '" + outputname
			      + "' not found!", __FILE__, __LINE__);
  outputprop[reg->index()].push_back(prop);
}

void Material::clear_outputprop() {
  for(std::vector<std::vector<Property*> >::size_type i=0; i<outputprop.size();
      i++)
    outputprop[i].clear();
};

const std::vector<Property*> &
Material::outputProperties(const PropertyOutput *pout) const {
  return outputprop[pout->index()];
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Element start/end.  Since each element has a unique material,
// these can only be called once per element.  The reason they're 
// here is because the Material is the one with the handy-dandy
// list of properties.  
void Material::begin_element(CSubProblem *subproblem, const Element *el) const {
//   Trace("Material::begin_element");
  for(std::vector<Property*>::size_type i=0;i<property.size();i++) {
    if(property[i]->currently_active(subproblem)) {
      property[i]->begin_element(subproblem, el);
    }
  }
}

void Material::end_element(CSubProblem *subproblem, const Element *el) const {
//   Trace("Material::end_element");
  for(std::vector<Property*>::size_type i=0;i<property.size();i++) {
    if(property[i]->currently_active(subproblem)) {
      property[i]->end_element(subproblem, el);
    }
  }
} 

void Material::precompute(CSubProblem *subproblem) {
  for(std::vector<Property*>::size_type i=0; i<property.size(); i++) {
    property[i]->find_active(subproblem);
    if(property[i]->currently_active(subproblem)) {
      property[i]->precompute(subproblem->mesh);
    }
  }
}

void Material::precompute_all_properties(FEMesh *mesh) {
  for(std::vector<Property*>::size_type i=0; i<property.size(); i++) {
    try {
      property[i]->precompute(mesh);
    }
    catch (ErrNoSuchField &exc) {
    }
  }
}

// The cross-referencing relies on properties not being shared
// between materials -- otherwise, the material is ambiguous in 
// precompute(), and in one of the contexts, the property can
// retrieve the wrong companion property.  Each material should
// have a unique copy  of each of its properties, and at this 
// point, a given property instance's precompute() routine should
// be called at most once.


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Post-processing is initially very simple -- just call each
// member property's post_process routine....
void Material::post_process(CSubProblem *subproblem, const Element *el) const {
  for(std::vector<Property*>::size_type i=0; i<property.size(); i++) {
    property[i]->post_process(subproblem, el);
  }
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// A Flux is active for this material if any of the Properties that
// contribute to it are active.  A Property is active if all of the
// Fields it uses are defined.

bool FluxPropList::active(const CSubProblem *subproblem) const {
  for(size_type i=0; i<size(); i++) {
    if((*this)[i]->currently_active(subproblem))
      return true;
  }
  return false;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Make a list of all fields that contribute to each flux.  Assumes
// that the FluxPropLists have already been constructed by
// registerFlux(), which is called by Property::bookkeeping().  Called
// by Material::cross_reference().

void Material::make_fluxfield_table() {
  // loop over all fluxes
//   Trace("Material::make_fluxfield_table");
  for(std::vector<FluxPropList>::size_type fluxi=0; fluxi<fluxprop.size();
      fluxi++)
    {
      std::vector<const Field*> &flist = fluxfield[fluxi];
      // loop over all properties that contribute to this flux
      FluxPropList &fpl = fluxprop[fluxi];
      for(FluxPropList::size_type propi=0; propi<fpl.size(); propi++) {
	Property *property = fpl[propi];
// 	Trace("Material::make_fluxfield_table: " + getFluxByIndex(fluxi)->name()
// 	      + " " + property->name())
	// loop over all fields required by this property
	const std::vector<const Field*> &fields = property->fields_required();
	for(std::vector<const Field*>::size_type fieldi=0; fieldi<fields.size();
	    fieldi++)
	  {
	    const Field *field = fields[fieldi];

	    // this field is needed by this property to compute this flux.
// 	    Trace("Material::make_fluxfield_table: needs "+field->name());
	    bool found = false;	// has it already been listed?
	    for(std::vector<const Field*>::size_type i=0;
		i<flist.size() && !found; i++)
	      {
		if(flist[i] == field) { // yes, it has
		  found = true;
		}
	      }
	    if(!found) {	// no, it hasn't already been listed
	      flist.push_back(field); // list it
	    }
	  }
      }
    }
}

void Material::clear_fluxfield_table() {
  for(std::vector<FluxPropList>::size_type fluxi=0; fluxi<fluxprop.size();
      fluxi++)
    fluxfield[fluxi].clear();
}

// Return a list of Fields relevant to the given Flux in this
// Material.  If the Flux isn't computed by any of this Material's
// Properties, then the list has no entries.

const std::vector<const Field*> &Material::fieldlist(const Flux &flux) const {
  return fluxfield[flux.index()];
}




//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=
//=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=

void Material::make_linear_system(CSubProblem *subproblem, 
				  const Element *el,
				  LinearizedSystem &system) const 
{
  //std::cerr << "Material::make_stiffness" << std::endl;
  // loop over all fluxes
//   Trace("Material::make_stiffness.");
  for(std::vector<FluxPropList>::size_type fluxi=0; fluxi<fluxprop.size();
      fluxi++)
    {
      Flux *flux = Flux::allfluxes()[fluxi];
//       Trace("Material::make_linear_system: " + flux->name());
      // fluxproplist is the list of Properties that contribute to this Flux
      const FluxPropList &fluxproplist = fluxprop[fluxi];

      // A Flux is active if it is used in one or more active Equations.
      // The fluxproplist is active if any of the Properties that
      // contribute to it are active.  A Property is active if all of
      // the Fields it uses are defined.
      if(subproblem->is_active_flux(*flux) && fluxproplist.active(subproblem)) {

	// Find the integration order for this flux.  Use the maximum
	// order required by the relevant Properties.  This could be
	// Element dependent, so it can't be precomputed.
	int order = 0;
	for(FluxPropList::size_type propi=0; propi<fluxproplist.size(); propi++)
	  {
	    int ord = fluxproplist[propi]->integration_order(subproblem, el);
	    if(ord > order) order = ord;
	  }
	order += flux->eqn_integration_order(subproblem, el);
	
      
	// Integrate over the Element
	for(GaussPointIterator gpt=el->integrator(order); !gpt.end(); ++gpt) {
// 	  Trace("Material::make_linear_system: gpt="
// 		+ to_string(gpt.gausspoint()));
	  FluxData *fluxdata = flux->initializeData(el);

	  // Loop over Properties contributing to this Flux,
	  // accumulating their contributions in the flux matrix. First,
	  // get the element's funcnode-iterator and pass it in. Do the
	  // iteration here, though, to ensure that each property in the
	  // fluxproplist gets a crack at each node.

	  for(FluxPropList::size_type propi=0; propi<fluxproplist.size();
	      propi++)
	    {
	      Property *prop = fluxproplist[propi];
// 	      Trace("Material::make_linear_system: property=" + prop->name());
	      if(prop->currently_active(subproblem)) {
		prop->begin_point(subproblem->mesh, el, flux, 
				  gpt.gausspoint());
		for(ElementFuncNodeIterator eni = el->funcnode_iterator();
		    !eni.end(); ++eni) {
// 		  Trace("Material::make_linear_system: node=" + to_string(eni));
		  // A flux matrix embodies a numerical representation 
		  // relating the nodal fluxes with its fields.

		  // For elastic systems:
		  // For the in-plane components, it is the "gradient", the 
		  // strain operator, \times the material stiffness,
		  // or equivalently the stiffness \times the nodal derivative 
		  // of the shape function contributions.
		  // For the out of plane components (plane stress), 
		  // it corresponds to the
		  // identity operator, or equivalently, to the stiffness tensor
		  // \times the nodal shape function.
		  // (NOT the derivative) 
	      
		  // For non-elastic systems: 
		  // For in-plane components, it is the "gradient" 
		  // operator \times the material parameter (e.g., thermal
		  // conductivity, mass diffusivity, etc.), or equivalently
		  // a proportionality constant \times the derivative of the
		  // shape function.
		  // For the out of plane components (plane flux) it corresponds
		  // to a proportionality constant times the shape function
		  // (NOT the derivative) 
// 		  std::cerr << "Material::make_linear_system: prop="
// 			    << fluxproplist[propi]->name() << std::endl;
		  fluxproplist[propi]->fluxmatrix(subproblem->mesh, el, eni,
						  flux, fluxdata,
						  gpt.gausspoint());
		}
		// For the RHS contributions, don't want to iterate over nodes.
		fluxproplist[propi]->fluxrhs(subproblem->mesh, el,
					     flux, fluxdata, gpt.gausspoint());
		prop->end_point(subproblem->mesh, el, flux, gpt.gausspoint());
		//fluxrhs embodies the body forces for each driving force
		//for mechanical systems could be gravity \times density,
		//For thermal systems, chemical reactions.
	      } //if "property is active" ends here
	    } //loop over all properties ends here

	  // Have the Flux make its contribution to the stiffness matrix
	  flux->make_stiffness(subproblem, el, gpt.gausspoint(), fluxdata,
			       system);
	  // ...and rhs.
	  flux->make_rhs(subproblem, el, gpt.gausspoint(), fluxdata, system);

	  delete fluxdata;
	} //loop over gauss points ends here
      } //if flux is active
    } //loop over all flux prop ends here
  
  /*
    TIME DERIVATIVES START HERE
  */
  if (subproblem->dynamics_activated() || subproblem->kinetics_activated())
    set_time_derivatives(subproblem, el, system);
  //time-dependent properties should be inserted at THIS level
  /*
    TIME DERIVATIVES END HERE
  */
} // make_linear_system



void Material::set_time_derivatives(CSubProblem *subproblem, 
				    const Element *el,
				    LinearizedSystem &system) const 
{
  const std::vector<Equation*>* eqns_list = subproblem->all_equations();
  for(std::vector<Equation*>::size_type i=0; i<eqns_list->size(); i++) {
    //for all equations, 	
    Equation * eqn = (*eqns_list)[i];
    const std::vector<TimeDerivative*> timeDervsList=eqn->allTimeDerivatives();
    for(std::vector<TimeDerivative*>::size_type j=0; j<timeDervsList.size();j++)
      {
	TimeDerivative * td = timeDervsList[j];
	// timeDerivproplist is the list of Properties that contribute
	// to this time derivative
	const std::vector<Property*> &timeDerivproplist =
	  td->get_properties(*this); 

	//start loop over all relevant properties and find highest
	//order of integration
	int order = 2*(el->shapefun_degree()); // <Unit> material assumes 2 times shape function degree
	// TODO CLEANUP: Why not just loop over properties, why indices?
	for(std::vector<Property*>::size_type propi=0;
	    propi<timeDerivproplist.size(); propi++)
	  {
	    int ord = timeDerivproplist[propi]->integration_order(subproblem,
								  el);
	    if(ord > order) 
	      order = ord;
	  } // end loop over all relevant properties and find highest
	    // order of integration
	order += eqn->integration_order(el);

	// For every gauss point, every time derivative of every
	// equation contributes to the c_matrix or m_matrix.  Here,
	// c_matrix holds all the first-order time derivatives
	// contributions, and m_matrix embodies all the second order
	// time-derivative contributions from every time-dependent
	// field.

	for(GaussPointIterator gpt=el->integrator(order); !gpt.end(); ++gpt) {
	  // depending on the order of the derivative, each
	  // TimeDerivative object contributes to its C- or M-matrix
	  td->make_matrix(subproblem, el, gpt.gausspoint(), system);
	} //loop over gauss points	
      } //loop over time derivatives of equation
  } //loop over all equations
}


// Find the flux matrix and offset for a single flux at a single
// point.  Used by Flux::evaluate.

void Material::find_fluxmatrix(const FEMesh *mesh, const Element *el,
			       const Flux *flux, FluxData *fluxdata,
			       const MasterPosition &mpos)
  const
{
  // List of Properties contributing to this flux.
  const FluxPropList &fluxproplist = fluxprop[flux->index()];
  for(FluxPropList::size_type propi=0; propi<fluxproplist.size(); propi++) {
    Property *property = fluxproplist[propi];
    // *Don't* check for active properties here, because the active_
    // flag is set correctly only when constructing the full stiffness
    // matrix.  find_fluxmatrix is called when computing a flux for
    // output.  Instead, rely on the exception thrown by
    // FEWrapper<Field>::offset when evaluating a Field at a Node.
    property->begin_point(mesh, el, flux, mpos);
    for(ElementFuncNodeIterator eni=el->funcnode_iterator(); !eni.end();++eni) {
      try {
	property->fluxmatrix(mesh, el, eni, flux, fluxdata, mpos);
      }
      catch (ErrNoSuchField &exc) {
	// Property can't be computed because some Fields aren't defined. 
      }
    }
    try {
      property->fluxrhs(mesh, el, flux, fluxdata, mpos);
    }
    catch (ErrNoSuchField &exc) {
    }
    property->end_point(mesh, el, flux, mpos);
  }
}

//=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=

// Returns a FluxPropList for a give Flux.
FluxPropList Material::get_flux_props(const Flux *fluks) const
{
  return fluxprop[fluks->index()];
}

#ifdef DEBUG
void Material::dump(std::ostream &os) const {
  os << "Material " << name() << ":" << std::endl;
  registry.dump(os);
}

#endif // DEBUG

//=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=

// Materials are assigned to Microstructure pixels with a
// PixelAttribute class.  See common/pixelattribute.h for details.

const std::string
MaterialAttributeRegistration::classname_("MaterialAttributeRegistration");
const std::string
MaterialAttributeRegistration::modulename_("ooflib.SWIG.engine.material");

static MaterialAttributeRegistration *matattrreg;

// The MaterialAttributeGlobalData class keeps track of how many
// pixels are assigned to each Material in a Microstructure.  This
// provides a quick way of determining which Materials are used in the
// Microstructure, which allows the Microstructure to respond to
// changes in Materials without looking through all of its pixels to
// see if the changed Material is present.

// MatMap counts how many times each Material is used in the
// Microstructure.
typedef std::map<const Material*, int, ltidobject> MatMap;

class MaterialAttributeGlobalData : public PixelAttributeGlobalData {
public:
  const CMicrostructure *microstructure;
  MatMap matmap;

  // getCount creats an entry for the material if it does not exist in
  // the map.  If you just want to query, use "count", below.
  int &getCount(const Material *mat) {
    MatMap::iterator i = matmap.find(mat);
    if(i == matmap.end()) {
      matmap[mat] = 0;
      return matmap[mat];
    }
    return (*i).second;
  }

  int count(const Material *mat) const {
    MatMap::const_iterator i = matmap.find(mat);
    if(i == matmap.end()) return 0;
    return (*i).second;
  }

  void remove(const Material *mat) {
    MatMap::iterator i = matmap.find(mat);
    if (i!=matmap.end()) 
      matmap.erase(i);
  }

  MaterialAttributeGlobalData(const CMicrostructure *microstructure)
    : microstructure(microstructure)
  {}
  void materialAddedToPixel(const Material *mat) {
    if(mat)
      getCount(mat)++;
  }
  void materialRemovedFromPixel(const Material *mat) {
    if(mat)
      getCount(mat)--;
  }

  std::vector<const Material*> *getMaterials() const {
    std::vector<const Material*> *mvec = new std::vector<const Material*>;
    for(MatMap::const_iterator i=matmap.begin(); i!=matmap.end(); ++i) 
      if((*i).second > 0) 
	mvec->push_back((*i).first);
    return mvec;
  }
  TimeStamp getTimeStamp() const {
    TimeStamp latest = timeZero;
    for(MatMap::const_iterator i=matmap.begin(); i!=matmap.end(); ++i) {
      const Material *mat = (*i).first;
      int count = (*i).second;
      if(count > 0) {
	const TimeStamp &ts = mat->getTimeStamp();
	if(ts > latest)
	  latest = ts;
      }
    }
    return latest;
  }
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

MaterialAttributeRegistration::MaterialAttributeRegistration()
  : PxlAttributeRegistration("Material")
{
  matattrreg = this;
}

PixelAttributeGlobalData *
MaterialAttributeRegistration::createAttributeGlobalData(const
							 CMicrostructure *ms)
  const
{
  return new MaterialAttributeGlobalData(ms);
}

void MaterialAttribute::set(const Material *m) {
  material = m;
}

// getMaterialMap and getConstMaterialMap are provided so that
// matattrreg can be a static variable in this file, but still make
// the map available to others.

Array<PixelAttribute*> &getMaterialMap(const CMicrostructure *ms) {
  return matattrreg->map(ms);
}

const Array<PixelAttribute*> &getConstMaterialMap(const CMicrostructure *ms) {
  return matattrreg->map(ms);
}
// Find all the MaterialAttributes in the given microstructure which
// refer to this material, and clear them.  If the material count is
// nonzero, iterate over all of of the MaterialAttributes and check
// them individually.

// TODO LATER: It is possible to be more clever about this, having the
// materials maintain a list of attributes which refer to them, and
// keeping it up to date, by tracking all MaterialAttribute set
// operations.  This is currently not done, because it imposes a large
// storage and searching cost on the materials themselves.  Searching
// through the whole array is probably slower, but it happens less
// frequently.

bool Material::cleanAttributes(CMicrostructure *ms) const {
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->globalData(ms));
#ifdef DEBUG
  assert(gd!=0);
#endif  
  if (gd->count(this)!=0) {
    const Array<PixelAttribute*> &mat_attrs = matattrreg->map(ms);
    for(Array<PixelAttribute*>::const_iterator i=mat_attrs.begin();
	i!=mat_attrs.end(); ++i) {
      MaterialAttribute *ma = dynamic_cast<MaterialAttribute*>(*i);
#ifdef DEBUG
      assert(ma!=0);
#endif
      if (this == ma->get()) 
	ma->set(0);
    }
    // Remove this material from the map, if it's present.
    gd->remove(this); 
    ms->recategorize();
    return true;
  }
  return false;
}

bool Material::replaceAttributes(CMicrostructure *ms, Material *newmat) const {
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->globalData(ms));
#ifdef DEBUG
  assert(gd!=0);
#endif  
  if (gd->count(this)!=0) {
    const Array<PixelAttribute*> &mat_attrs = matattrreg->map(ms);
    for(Array<PixelAttribute*>::const_iterator i=mat_attrs.begin();
	i!=mat_attrs.end(); ++i) {
      MaterialAttribute *ma = dynamic_cast<MaterialAttribute*>(*i);
#ifdef DEBUG
      assert(ma!=0);
#endif
      if (this == ma->get()) {
	ma->set(newmat);  
	gd->materialAddedToPixel(newmat);
      }
    }
    // Remove this material from the map, if it's present.
    gd->remove(this); 
    ms->recategorize();
    return true;
  }
  return false;
}

bool MaterialAttribute::operator<(const PixelAttribute &other) const {
  const MaterialAttribute &othermat =
    dynamic_cast<const MaterialAttribute&>(other);
  if(material != 0 && othermat.material != 0)
    return material->name() < othermat.material->name();
  if(material == 0 && othermat.material != 0)
    return true;
  return false;
}

void MaterialAttribute::print(std::ostream &os) const {
  if(material)
    os << "Material(" << material->name() << ")";
  else
    os << "[No Material]";
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Material::assignToPixels(CMicrostructure *microstructure,
			      const std::vector<ICoord> *pxls) const
{
  // First, remove the old material, if any.
  removeMaterialFromPixels(microstructure, *pxls);

  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  const ActiveArea *activearea = microstructure->getActiveArea();
  for(std::vector<ICoord>::const_iterator i=pxls->begin(); i<pxls->end(); ++i) {
    if(activearea->isActive(*i)) {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(matMap[*i]);
      matAtt->set(this);
      gd->materialAddedToPixel(this);
    }
  }
  microstructure->recategorize();
}

void Material::assignToPixelGroup(CMicrostructure *microstructure,
				  const PixelSet *pixset) const
{
  // Assign material to the pixels in the given group.
  assignToPixels(microstructure, pixset->members());
}

void Material::assignToAllPixels(CMicrostructure *microstructure) const {
  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  const ActiveArea *activearea = microstructure->getActiveArea();
  for(Array<PixelAttribute*>::iterator i=matMap.begin(); i!=matMap.end(); ++i) {
    if(activearea->isActive(i.coord())) {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(*i);
#ifdef DEBUG
      assert(matAtt!=0);
#endif
      gd->materialRemovedFromPixel(matAtt->get());
      matAtt->set(this);
      gd->materialAddedToPixel(this);
    }
  }
  microstructure->recategorize();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

int Material::nPixelsInMicrostructure(const CMicrostructure *microstructure)
  const
{
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  return gd->count(this);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void removeMaterialFromPixels(CMicrostructure *microstructure,
			      const std::vector<ICoord> &pxls)
{
  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  const ActiveArea *activearea = microstructure->getActiveArea();
  for(std::vector<ICoord>::const_iterator i=pxls.begin(); i<pxls.end(); ++i) {
    if(activearea->isActive(*i)) {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(matMap[*i]);
#ifdef DEBUG
      assert(matAtt!=0);
#endif
      gd->materialRemovedFromPixel(matAtt->get());
      matAtt->set(0);
    }
  }
  microstructure->recategorize();
}

void removeMaterialFromPixels(CMicrostructure *microstructure,
				const PixelSet *pixset)
{
  removeMaterialFromPixels(microstructure, *pixset->members());
}

void removeAllMaterials(CMicrostructure *microstructure) {
  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  const ActiveArea *activearea = microstructure->getActiveArea();
  for(Array<PixelAttribute*>::iterator i=matMap.begin(); i!=matMap.end(); ++i) {
    if(activearea->isActive(i.coord())) {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(*i);
#ifdef DEBUG
      assert(matAtt!=0);
#endif
      gd->materialRemovedFromPixel(matAtt->get());
      matAtt->set(0);
    }
  }
  microstructure->recategorize();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

TimeStamp getMaterialTimeStamp(const CMicrostructure *microstructure) {
  MaterialAttributeGlobalData *gd =
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  return gd->getTimeStamp();
}

const Material *getMaterialFromCategory(const CMicrostructure *microstructure,
					int category)
{
  const ICoord &where = microstructure->getRepresentativePixel(category);
  return getMaterialFromPoint(microstructure, &where);
}

const Material *getMaterialFromPoint(const CMicrostructure *microstructure,
				     const ICoord *where)
{
  Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(matMap[*where]);
#ifdef DEBUG
  assert(matAtt!=0);
#endif
  return matAtt->get();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

MaterialImage::MaterialImage(CMicrostructure *microstructure,
			     const CColor *noMaterial,
			     const CColor *noColor)
  : microstructure(microstructure),
    noMaterial(*noMaterial),
    noColor(*noColor)
{}

const Coord &MaterialImage::size() const {
  return microstructure->size();
}

const ICoord &MaterialImage::sizeInPixels() const {
  return microstructure->sizeInPixels();
}

void MaterialImage::fillstringimage(StringImage *strimg) const {
  const Array<PixelAttribute*> &matMap = matattrreg->map(microstructure);
  for(Array<PixelAttribute*>::const_iterator i=matMap.begin(); i!=matMap.end();
      ++i)
    {
      MaterialAttribute *matAtt = dynamic_cast<MaterialAttribute*>(*i);
      const Material *mat = matAtt->get();
      if(mat) {
	try {
	  Property *cprop = mat->fetchProperty("Color");
	  ColorProp *colorprop = dynamic_cast<ColorProp*>(cprop);
	  const CColor &cc = colorprop->color();
	  strimg->set(&i.coord(), &cc);
	}
	catch(ErrNoSuchProperty &exc) {
	  strimg->set(&i.coord(), &noColor);
	}
      }	// if mat
      else
	strimg->set(&i.coord(), &noMaterial);
    }
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Return a list of all Materials used in a Microstructure.
std::vector<const Material*> *
getMaterials(const CMicrostructure *microstructure) {
  MaterialAttributeGlobalData *gd = 
    dynamic_cast<MaterialAttributeGlobalData*>(matattrreg->
					       globalData(microstructure));
#ifdef DEBUG
  assert(gd!=0);
#endif
  return gd->getMaterials();
}
